#include <xmms/plugin.h>
#include <SDL/SDL.h>
#include <list>
#include <string>

#define DEBUG_STRING string(__FILE__) + ":" + Misc::int2string(__LINE__) + ":" + \
      __PRETTY_FUNCTION__ + ": "

namespace RAHXMMSVIS {

void * xmalloc (size_t size);
void * xrealloc (void * mem, size_t size);

class Misc {
public:
  static string int2string (int);
  static int randomNumber (double, double);
};

typedef enum { Pulse,
               Surface } DataType;

class Exception {
public:
  string _desc;
  Exception (const string& desc) : _desc (desc) {}
};

class Data {
public:
  DataType _type;
  Data (DataType type)
  : _type (type) {}
  virtual ~Data () {}
};

class PulseData : public Data {
protected:
  int _value;
public:
  PulseData (int value)
  : Data (Pulse),
    _value (value) {}
  PulseData (const PulseData * pulseData)
  : Data (Pulse), _value (pulseData->_value) {}
  int getValue () {
    return _value;
  }
  void setValue (int value) {
    _value = value;
  }
  virtual ~PulseData () {}
};

class SurfaceData : public Data {
protected: 
  SDL_Surface * _surface;
  bool _created;
public:
  SurfaceData (SDL_Surface * surface)
  : Data (Surface),
    _surface (surface),
    _created (false) {}
  SurfaceData (const SDL_Surface * surface)
  : Data (Surface),
    _surface (SDL_CreateRGBSurface(SDL_HWSURFACE,
                                   surface->w,
                                   surface->h,
                                   surface->format->BitsPerPixel,
                                   surface->format->Rmask,
                                   surface->format->Gmask,
                                   surface->format->Bmask,
                                   surface->format->Amask)),
    _created (true) {
    if (!_surface) {
      string s (DEBUG_STRING + "could not create surface: " + SDL_GetError());
      cerr << s << endl;
      throw Exception (s);
    }
    SDL_BlitSurface ((SDL_Surface *) surface, NULL, _surface, NULL);
    SDL_UpdateRect((SDL_Surface *) surface, 0, 0, surface->w, surface->h);
  }
  SDL_Surface * getSurface () {
    return _surface;
  }
  virtual ~SurfaceData () {
    if (_created && _surface) SDL_FreeSurface (_surface);
  }
};

class Node;

class Port {
protected:
  Node * _node;
public:
  DataType _type;
  Port (DataType type, Node * node)
  : _type (type), _node (node) {}
  virtual void process () = 0;
  Node * node () { return _node; }
};

class Socket;

class Plug : public Port {
private:
  list<Socket *> _connectedSockets;
public:
  Plug (DataType type, Node * node)
  : Port (type, node) {}
  void plugInto (Socket * socket);
  void write (Data * data);
  void unplugAllFrom (Socket * socket);
  virtual void process ();
};

class Socket : public Port {
private:
  list<Data *> _dataBuffer;
  Plug * _connectedPlug;
public:
  Socket (DataType type, Node * node)
  : Port (type, node) {}
  void write (Data *);
  Data * read ();
  virtual void process ();
  bool hasData () {
    return !_dataBuffer.empty();
  }
  void connect (Plug * plug) {
    _connectedPlug = plug;
  }
  void unconnect () {
    _connectedPlug = 0;
  }
};

/*class Port {
private:
  list<Data *> _dataBuffer;
  list<Port *> _connectedPorts;
  Node * _node;
public:
  DataType _type;
  Port (DataType type, Node * node)
  : _type (type),
    _node (node) {}
  /// doesn't copy data
/*  Port (const Port& port)
  : _connectedPort (port._otherPort),
    _node (port._node),
    _type (port._type) {}*//*
  Data * read ();
  void write (Data * data);
  void plugInto (Port * port);
  void unplugAllFrom (Port * port);
  void process ();
  bool hasData () {
    return !_dataBuffer.empty();
  }
};*/

class Node {
private:
public:
  list<Plug *> plugs;
  list<Socket *> sockets;
  virtual void process() = 0;
  virtual ~Node () {};
  void connectPortRanges (list<Plug *>::iterator& plug,
                          list<Socket *>::iterator& socket,
                          unsigned count);
  bool allSocketsHaveData ();
friend Port;
};


/**
 * SDL Surface sink.
 * Sockets: 0 - Surface
 * Plugs: none
 */
class SDLNode : public Node {
private:
  int _numberOfScreens;
  list<SDL_Surface *> _surfaces;
public:
  SDLNode ();
  virtual void process ();
};


/**
 * XMMS frequency data.
 * Sockets: none
 * Plugs: 0-255 - Pulse
 */
class XmmsNode : public Node {
private:
  list<int> _freqData;
public:
  XmmsNode ();
  void setData (list<int>& data) {
    _freqData = data;
  }
  virtual void process ();
};

/**
 * Holds a bitmap.
 * Sockets: none
 * Plugs: 0 - Surface
 */
class BitmapNode : public Node {
private:
  SDL_Surface * _surface;
  unsigned _plugCount;
public:
  BitmapNode (const string& filename, unsigned plugCount) 
   : _surface (SDL_LoadBMP(filename.c_str())),
     _plugCount (plugCount) {
    if (!_surface) {
      cerr << DEBUG_STRING + "could load bitmap from \"" + filename +
        "\":" + string(SDL_GetError());
    }
    for (int i = 0; i < plugCount; ++i) {
      plugs.push_back(new Plug (Surface, this));
    }
  }
  virtual void process () {
    cerr << DEBUG_STRING + "begin" <<  endl;
    for (list<Plug *>::iterator itp = plugs.begin ();
         itp != plugs.end ();
         ++itp) {
      (*itp)->write (new SurfaceData ((const SDL_Surface *) _surface));
    }
    for (list<Plug *>::iterator itp = plugs.begin ();
         itp != plugs.end ();
         ++itp) {
      (*itp)->process ();
    }
    cerr << DEBUG_STRING + "end" <<  endl;
  }
  virtual ~BitmapNode () {
    SDL_FreeSurface (_surface);
  }
};

/**
 * Averages pulse data between equal-sized partitions.
 * Sockets: 0 to inputs - Pulse
 * Plugs: 0 to (2*bands) - Pulse
 */
class EQNode : public Node {
private:
  unsigned _inputs;
  unsigned _bands;
public:
  EQNode (unsigned inputs, unsigned bands)
   : _inputs (inputs),
     _bands (bands) {
    int i;
    for (i = 0; i < inputs; ++i) {
      sockets.push_back(new Socket (Pulse, this));
    }
    for (i = 0; i < bands*2; ++i) {
      plugs.push_back(new Plug (Pulse, this));
    }
  }
  virtual ~EQNode () {}
  virtual void process ();
};

/**
 * Flickers an image according to a pulse.
 * Sockets: 0 - Pulse, 1 - Surface (image to flicker), 2 - Surface (background to flicker on)
 * Plugs: 0 - Surface
 */
class ImageFlickerNode : public Node {
private:
public:
  ImageFlickerNode () {
    sockets.push_back (new Socket (Pulse, this));
    sockets.push_back (new Socket (Surface, this)); // image to flicker
    sockets.push_back (new Socket (Surface, this)); // background
    plugs.push_back (new Plug (Surface, this));
  }
  virtual ~ImageFlickerNode () {}
  virtual void process ();
};

/**
 * Overlays one image with a number of other images.
 * Sockets: 0 - Surface (background), n - n+1 (for each odd n >= 1):
 *          n - Pulse (alpha for surface),
 *          n+1 - Surface
 * Plugs: 0 - Surface
 */
class ImageOverlayNode : public Node {
  unsigned _imageCount;
public:
  ImageOverlayNode (unsigned imageCount)
  : _imageCount (imageCount) {
    plugs.push_back (new Plug (Surface, this));
    sockets.push_back (new Socket (Surface, this));
    for (int i = 0; i < imageCount; ++i) {
      sockets.push_back (new Socket (Pulse, this));
      sockets.push_back (new Socket (Surface, this));
    }
  }
  virtual ~ImageOverlayNode () {}
  virtual void process ();
};




} /* namespace RAHXMMSVIS */
